@@ -17,7 +17,7 @@
  * License along with this library; if not, write to the Free Software
  * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  * 
- * $Id: dbi_main.c,v 1.56 2004/01/03 17:23:54 mhoenicka Exp $
+ * $Id: dbi_main.c,v 1.75 2006/06/02 17:18:16 qu1j0t3 Exp $
  */
 
 #ifdef HAVE_CONFIG_H
@@ -38,6 +38,7 @@
 #define my_dlsym lt_dlsym
 #define my_dlclose lt_dlclose
 #define my_dlerror lt_dlerror
+
 #elif HAVE_MACH_O_DYLD_H
 #include <mach-o/dyld.h>
 static void * dyld_dlopen(const char * file);
@@ -48,12 +49,29 @@ static const char * dyld_dlerror();
 #define my_dlsym dyld_dlsym
 #define my_dlerror dyld_dlerror
 #define my_dlclose dyld_dlclose
+
+#elif __MINGW32__
+#include <windows.h>
+void *win_dlopen(const char*, int);
+void *win_dlsym(void *, const char*);
+int win_dlclose(void *);
+char *win_dlerror();
+/* just for compiling support,if anyone has used these masks in code. The MODE argument to `dlopen' contains one of the following: */
+#define RTLD_LAZY       0x001   /* Lazy function call binding.  */
+#define RTLD_NOW        0x002   /* Immediate function call binding.  */
+#define RTLD_BINDING_MASK 0x003   /* Mask of binding time value.  */
+#define my_dlopen(foo,bar) win_dlopen(foo,bar)
+#define my_dlsym win_dlsym
+#define my_dlclose win_dlclose
+#define my_dlerror win_dlerror
+
 #elif HAVE_DLFCN_H
 #include <dlfcn.h>
 #define my_dlopen dlopen
 #define my_dlsym dlsym
 #define my_dlclose dlclose
 #define my_dlerror dlerror
+
 #else
 #error no dynamic loading support
 #endif
@@ -69,14 +87,31 @@ static const char * dyld_dlerror();
 #include <dbi/dbi.h>
 #include <dbi/dbi-dev.h>
 
-#ifndef DBI_DRIVER_DIR
-#define DBI_DRIVER_DIR "/usr/local/lib/dbd" /* use this as the default */
+#ifdef __MINGW32__
+#  define DBI_PATH_SEPARATOR "\\"
+#  ifndef DBI_DRIVER_DIR
+#	define DBI_DRIVER_DIR "c:\\libdbi\\lib\\dbd" /* use this as the default */
+#  endif
+#  else
+#    define DBI_PATH_SEPARATOR "/"
+#    ifndef DBI_DRIVER_DIR
+#    	define DBI_DRIVER_DIR "/usr/local/lib/dbd" /* use this as the default */
+#    endif
 #endif
 
 #ifndef DLSYM_PREFIX
 #define DLSYM_PREFIX ""
 #endif
 
+/* declarations of optional external functions */
+#ifndef HAVE_VASPRINTF
+int int_vasprintf(char **result, const char *format, va_list *args);
+int vasprintf(char **result, const char *format, va_list args);
+#endif
+#ifndef HAVE_ASPRINTF
+int asprintf(char **result, const char *format, ...);
+#endif
+
 /* declarations for internal functions -- anything declared as static won't be accessible by name from client programs */
 static dbi_driver_t *_get_driver(const char *filename);
 static void _free_custom_functions(dbi_driver_t *driver);
@@ -85,6 +120,7 @@ static int _update_internal_conn_list(dbi_conn_t *conn, int operation);
 static void _free_caps(_capability_t *caproot);
 static const char *_get_option(dbi_conn Conn, const char *key, int aggressive);
 static int _get_option_numeric(dbi_conn Conn, const char *key, int aggressive);
+static unsigned int _parse_versioninfo(const char *version);
 
 void _error_handler(dbi_conn_t *conn, dbi_error_flag errflag);
 extern int _disjoin_from_conn(dbi_result_t *result);
@@ -92,10 +128,11 @@ extern int _disjoin_from_conn(dbi_result_t *result);
 dbi_result dbi_conn_queryf(dbi_conn Conn, const char *formatstr, ...) __attribute__ ((format (printf, 2, 3)));
 int dbi_conn_set_error(dbi_conn Conn, int errnum, const char *formatstr, ...) __attribute__ ((format (printf, 3, 4)));
 
-static const char *ERROR = "ERROR";
+/* must not be called "ERROR" due to a name clash on Windoze */
+static const char *my_ERROR = "ERROR";
 static dbi_driver_t *rootdriver;
 static dbi_conn_t *rootconn;
-static int complain = 1;
+int dbi_verbosity = 1;
 
 /* XXX DBI CORE FUNCTIONS XXX */
 
@@ -122,7 +159,7 @@ int dbi_initialize(const char *driverdir) {
 	else {
 		while ((driver_dirent = readdir(dir)) != NULL) {
 			driver = NULL;
-			snprintf(fullpath, FILENAME_MAX, "%s/%s", effective_driverdir, driver_dirent->d_name);
+			snprintf(fullpath, FILENAME_MAX, "%s%s%s", effective_driverdir, DBI_PATH_SEPARATOR, driver_dirent->d_name);
 			if ((stat(fullpath, &statbuf) == 0) && S_ISREG(statbuf.st_mode) && strrchr(driver_dirent->d_name, '.') && (!strcmp(strrchr(driver_dirent->d_name, '.'), DRIVER_EXT))) {
 				/* file is a stat'able regular file that ends in .so (or appropriate dynamic library extension) */
 				driver = _get_driver(fullpath);
@@ -141,7 +178,7 @@ int dbi_initialize(const char *driverdir) {
 					if (driver && driver->functions) free(driver->functions);
 					if (driver) free(driver);
 					driver = NULL; /* don't include in linked list */
-					if (complain) fprintf(stderr, "libdbi: Failed to load driver: %s\n", fullpath);
+					if (dbi_verbosity) fprintf(stderr, "libdbi: Failed to load driver: %s\n", fullpath);
 				}
 			}
 		}
@@ -189,8 +226,8 @@ int dbi_set_verbosity(int verbosity) {
 	 * isn't handled through the normal connection-oriented DBI error
 	 * mechanisms) */
 
-	int prev = complain;
-	complain = verbosity;
+	int prev = dbi_verbosity;
+	dbi_verbosity = verbosity;
 	return prev;
 }
 
@@ -281,7 +318,7 @@ int dbi_conn_cap_get(dbi_conn Conn, const char *capname) {
 const char *dbi_driver_get_name(dbi_driver Driver) {
 	dbi_driver_t *driver = Driver;
 	
-	if (!driver) return ERROR;
+	if (!driver) return my_ERROR;
 	
 	return driver->info->name;
 }
@@ -289,7 +326,7 @@ const char *dbi_driver_get_name(dbi_driver Driver) {
 const char *dbi_driver_get_filename(dbi_driver Driver) {
 	dbi_driver_t *driver = Driver;
 	
-	if (!driver) return ERROR;
+	if (!driver) return my_ERROR;
 	
 	return driver->filename;
 }
@@ -297,7 +334,7 @@ const char *dbi_driver_get_filename(dbi_driver Driver) {
 const char *dbi_driver_get_description(dbi_driver Driver) {
 	dbi_driver_t *driver = Driver;
 	
-	if (!driver) return ERROR;
+	if (!driver) return my_ERROR;
 	
 	return driver->info->description;
 }
@@ -305,7 +342,7 @@ const char *dbi_driver_get_description(dbi_driver Driver) {
 const char *dbi_driver_get_maintainer(dbi_driver Driver) {
 	dbi_driver_t *driver = Driver;
 	
-	if (!driver) return ERROR;
+	if (!driver) return my_ERROR;
 
 	return driver->info->maintainer;
 }
@@ -313,7 +350,7 @@ const char *dbi_driver_get_maintainer(dbi_driver Driver) {
 const char *dbi_driver_get_url(dbi_driver Driver) {
 	dbi_driver_t *driver = Driver;
 
-	if (!driver) return ERROR;
+	if (!driver) return my_ERROR;
 
 	return driver->info->url;
 }
@@ -321,7 +358,7 @@ const char *dbi_driver_get_url(dbi_driver Driver) {
 const char *dbi_driver_get_version(dbi_driver Driver) {
 	dbi_driver_t *driver = Driver;
 	
-	if (!driver) return ERROR;
+	if (!driver) return my_ERROR;
 	
 	return driver->info->version;
 }
@@ -329,28 +366,28 @@ const char *dbi_driver_get_version(dbi_driver Driver) {
 const char *dbi_driver_get_date_compiled(dbi_driver Driver) {
 	dbi_driver_t *driver = Driver;
 	
-	if (!driver) return ERROR;
+	if (!driver) return my_ERROR;
 	
 	return driver->info->date_compiled;
 }
 
-int dbi_driver_quote_string_copy(dbi_driver Driver, const char *orig, char **newquoted) {
+size_t dbi_driver_quote_string_copy(dbi_driver Driver, const char *orig, char **newquoted) {
 	dbi_driver_t *driver = Driver;
 	char *newstr;
-	int newlen;
+	size_t newlen;
 	
-	if (!driver || !orig || !newquoted) return -1;
+	if (!driver || !orig || !newquoted) return 0;
 
 	newstr = malloc((strlen(orig)*2)+4+1); /* worst case, we have to escape every character and add 2*2 surrounding quotes */
 
 	if (!newstr) {
-		return -1;
+		return 0;
 	}
 	
 	newlen = driver->functions->quote_string(driver, orig, newstr);
-	if (newlen < 0) {
+	if (!newlen) {
 		free(newstr);
-		return -1;
+		return 0;
 	}
 
 	*newquoted = newstr;
@@ -358,16 +395,21 @@ int dbi_driver_quote_string_copy(dbi_driver Driver, const char *orig, char **new
 	return newlen;
 }
 
-int dbi_driver_quote_string(dbi_driver Driver, char **orig) {
+size_t dbi_driver_quote_string(dbi_driver Driver, char **orig) {
 	char *temp = NULL;
 	char *newstr = NULL;
-	int newlen;
+	size_t newlen;
 
 	if (!orig || !*orig) {
-		return -1;
+		return 0;
 	}
 	
 	newlen = dbi_driver_quote_string_copy(Driver, *orig, &newstr);
+	if (!newlen) {
+	  /* in case of an error, leave the original string alone */
+	  return 0;
+	}
+
 	temp = *orig;
 	*orig = newstr;
 	free(temp); /* original unescaped string */
@@ -375,6 +417,27 @@ int dbi_driver_quote_string(dbi_driver Driver, char **orig) {
 	return newlen;
 }
 
+const char* dbi_driver_encoding_from_iana(dbi_driver Driver, const char* iana_encoding) {
+  dbi_driver_t *driver = Driver;
+	
+  if (!driver) {
+    return NULL;
+  }
+
+  return driver->functions->encoding_from_iana(iana_encoding);
+}
+
+const char* dbi_driver_encoding_to_iana(dbi_driver Driver, const char* db_encoding) {
+  dbi_driver_t *driver = Driver;
+	
+  if (!driver) {
+    return NULL;
+  }
+
+  return driver->functions->encoding_to_iana(db_encoding);
+}
+
+
 /* XXX DRIVER FUNCTIONS XXX */
 
 dbi_conn dbi_conn_new(const char *name) {
@@ -395,7 +458,7 @@ dbi_conn dbi_conn_open(dbi_driver Driver) {
 		return NULL;
 	}
 
-	conn = (dbi_conn_t *) malloc(sizeof(dbi_conn_t));
+	conn = malloc(sizeof(dbi_conn_t));
 	if (!conn) {
 		return NULL;
 	}
@@ -454,6 +517,7 @@ void dbi_conn_close(dbi_conn Conn) {
 	free(conn->results);
 
 	free(conn);
+
 }
 
 dbi_driver dbi_conn_get_driver(dbi_conn Conn) {
@@ -469,17 +533,19 @@ int dbi_conn_error(dbi_conn Conn, const char **errmsg_dest) {
 	char number_portion[20];
 	static char *errmsg = NULL; // XXX quick hack, revisit this when API is redesigned
 
-	if (errmsg) free(errmsg);
-	
-	if (conn->error_number) {
-		snprintf(number_portion, 20, "%d: ", conn->error_number);
-	}
-	else {
-		number_portion[0] = '\0';
-	}
+	if (errmsg_dest) {
+		if (errmsg) free(errmsg);
+		
+		if (conn->error_number) {
+			snprintf(number_portion, 20, "%d: ", conn->error_number);
+		}
+		else {
+			number_portion[0] = '\0';
+		}
 
-	asprintf(&errmsg, "%s%s", number_portion, conn->error_message ? conn->error_message : "");
-	*errmsg_dest = errmsg;
+		asprintf(&errmsg, "%s%s", number_portion, conn->error_message ? conn->error_message : "");
+		*errmsg_dest = errmsg;
+	}
 
 	return conn->error_number;
 }
@@ -528,6 +594,70 @@ int dbi_conn_set_error(dbi_conn Conn, int errnum, const char *formatstr, ...) {
 	return len;
 }
 
+size_t dbi_conn_quote_string_copy(dbi_conn Conn, const char *orig, char **newquoted) {
+	dbi_conn_t *conn = Conn;
+	char *newstr;
+	size_t newlen;
+	
+	if (!Conn || !orig || !newquoted) return 0;
+
+	newstr = malloc((strlen(orig)*2)+4+1); /* worst case, we have to escape every character and add 2*2 surrounding quotes */
+
+	if (!newstr) {
+		return 0;
+	}
+	
+	newlen = conn->driver->functions->conn_quote_string(conn, orig, newstr);
+	if (!newlen) {
+		free(newstr);
+		return 0;
+	}
+
+	*newquoted = newstr;
+
+	return newlen;
+}
+
+size_t dbi_conn_quote_string(dbi_conn Conn, char **orig) {
+	char *temp = NULL;
+	char *newstr = NULL;
+	size_t newlen;
+
+	if (!orig || !*orig) {
+		return 0;
+	}
+	
+	newlen = dbi_conn_quote_string_copy(Conn, *orig, &newstr);
+	if (!newlen) {
+	  /* leave original string alone in case of an error */
+	  return 0;
+	}
+	temp = *orig;
+	*orig = newstr;
+	free(temp); /* original unescaped string */
+
+	return newlen;
+}
+
+size_t dbi_conn_quote_binary_copy(dbi_conn Conn, const unsigned char *orig, size_t from_length, unsigned char **ptr_dest) {
+  unsigned char *temp = NULL;
+  size_t newlen;
+  dbi_conn_t *conn = Conn;
+
+  if (!Conn || !orig || !ptr_dest) {
+    return 0;
+  }
+
+  newlen = conn->driver->functions->quote_binary(conn, orig, from_length, &temp);
+  if (!newlen) {
+    return 0;
+  }
+
+  *ptr_dest = temp;
+
+  return newlen;
+}
+
 /* DRIVER: option manipulation */
 
 int dbi_conn_set_option(dbi_conn Conn, const char *key, const char *value) {
@@ -650,7 +780,7 @@ const char *dbi_conn_get_option_list(dbi_conn Conn, const char *current) {
 
 void dbi_conn_clear_option(dbi_conn Conn, const char *key) {
 	dbi_conn_t *conn = Conn;
-	dbi_option_t *prevoption;
+	dbi_option_t *prevoption = NULL; /* shut up compiler */
 	dbi_option_t *option;
 	
 	if (!conn) return;
@@ -734,6 +864,25 @@ const char *dbi_conn_get_encoding(dbi_conn Conn){
 	return retval;
 }
 
+unsigned int dbi_conn_get_engine_version(dbi_conn Conn){
+	dbi_conn_t *conn = Conn;
+	char versionstring[VERSIONSTRING_LENGTH];
+
+	if (!conn) return 0;
+
+	conn->driver->functions->get_engine_version(conn, versionstring);
+
+	return _parse_versioninfo(versionstring);
+}
+
+char* dbi_conn_get_engine_version_string(dbi_conn Conn, char *versionstring) {
+	dbi_conn_t *conn = Conn;
+
+	if (!conn) return 0;
+
+	return conn->driver->functions->get_engine_version(conn, versionstring);
+}
+
 dbi_result dbi_conn_get_db_list(dbi_conn Conn, const char *pattern) {
 	dbi_conn_t *conn = Conn;
 	dbi_result_t *result;
@@ -770,6 +919,7 @@ dbi_result dbi_conn_query(dbi_conn Conn, const char *statement) {
 
 	if (!conn) return NULL;
 	
+	_logquery(conn, "[query] %s\n", statement);
 	result = conn->driver->functions->query(conn, statement);
 
 	if (result == NULL) {
@@ -791,6 +941,7 @@ dbi_result dbi_conn_queryf(dbi_conn Conn, const char *formatstr, ...) {
 	vasprintf(&statement, formatstr, ap);
 	va_end(ap);
 	
+	_logquery(conn, "[queryf] %s\n", statement);
 	result = conn->driver->functions->query(conn, statement);
 
 	if (result == NULL) {
@@ -801,12 +952,13 @@ dbi_result dbi_conn_queryf(dbi_conn Conn, const char *formatstr, ...) {
 	return (dbi_result)result;
 }
 
-dbi_result dbi_conn_query_null(dbi_conn Conn, const unsigned char *statement, unsigned long st_length) {
+dbi_result dbi_conn_query_null(dbi_conn Conn, const unsigned char *statement, size_t st_length) {
 	dbi_conn_t *conn = Conn;
 	dbi_result_t *result;
 
 	if (!conn) return NULL;
 
+	_logquery_null(conn, statement, st_length);
 	result = conn->driver->functions->query_null(conn, statement, st_length);
 
 	if (result == NULL) {
@@ -818,7 +970,7 @@ dbi_result dbi_conn_query_null(dbi_conn Conn, const unsigned char *statement, un
 
 int dbi_conn_select_db(dbi_conn Conn, const char *db) {
 	dbi_conn_t *conn = Conn;
-	char *retval;
+	const char *retval;
 	
 	if (!conn) return -1;
 	
@@ -883,17 +1035,18 @@ static dbi_driver_t *_get_driver(const char *filename) {
 	dlhandle = my_dlopen(filename, DLOPEN_FLAG); /* DLOPEN_FLAG defined by autoconf */
 
 	if (dlhandle == NULL) {
+	  fprintf(stderr, "%s\n", my_dlerror());
 		return NULL;
 	}
 	else {
-		driver = (dbi_driver_t *) malloc(sizeof(dbi_driver_t));
+		driver = malloc(sizeof(dbi_driver_t));
 		if (!driver) return NULL;
 
 		driver->dlhandle = dlhandle;
 		driver->filename = strdup(filename);
 		driver->next = NULL;
 		driver->caps = NULL;
-		driver->functions = (dbi_functions_t *) malloc(sizeof(dbi_functions_t));
+		driver->functions = malloc(sizeof(dbi_functions_t));
 
 		if ( /* nasty looking if block... is there a better way to do it? */
 			((driver->functions->register_driver = my_dlsym(dlhandle, DLSYM_PREFIX "dbd_register_driver")) == NULL) ||
@@ -905,11 +1058,16 @@ static dbi_driver_t *_get_driver(const char *filename) {
 			((driver->functions->goto_row = my_dlsym(dlhandle, DLSYM_PREFIX "dbd_goto_row")) == NULL) ||
 			((driver->functions->get_socket = my_dlsym(dlhandle, DLSYM_PREFIX "dbd_get_socket")) == NULL) ||
 			((driver->functions->get_encoding = my_dlsym(dlhandle, DLSYM_PREFIX "dbd_get_encoding")) == NULL) ||
+			((driver->functions->encoding_from_iana = my_dlsym(dlhandle, DLSYM_PREFIX "dbd_encoding_from_iana")) == NULL) ||
+			((driver->functions->encoding_to_iana = my_dlsym(dlhandle, DLSYM_PREFIX "dbd_encoding_to_iana")) == NULL) ||
+			((driver->functions->get_engine_version = my_dlsym(dlhandle, DLSYM_PREFIX "dbd_get_engine_version")) == NULL) ||
 			((driver->functions->list_dbs = my_dlsym(dlhandle, DLSYM_PREFIX "dbd_list_dbs")) == NULL) ||
 			((driver->functions->list_tables = my_dlsym(dlhandle, DLSYM_PREFIX "dbd_list_tables")) == NULL) ||
 			((driver->functions->query = my_dlsym(dlhandle, DLSYM_PREFIX "dbd_query")) == NULL) ||
 			((driver->functions->query_null = my_dlsym(dlhandle, DLSYM_PREFIX "dbd_query_null")) == NULL) ||
 			((driver->functions->quote_string = my_dlsym(dlhandle, DLSYM_PREFIX "dbd_quote_string")) == NULL) ||
+			((driver->functions->quote_binary = my_dlsym(dlhandle, DLSYM_PREFIX "dbd_quote_binary")) == NULL) ||
+			((driver->functions->conn_quote_string = my_dlsym(dlhandle, DLSYM_PREFIX "dbd_conn_quote_string")) == NULL) ||
 			((driver->functions->select_db = my_dlsym(dlhandle, DLSYM_PREFIX "dbd_select_db")) == NULL) ||
 			((driver->functions->geterror = my_dlsym(dlhandle, DLSYM_PREFIX "dbd_geterror")) == NULL) ||
 			((driver->functions->get_seq_last = my_dlsym(dlhandle, DLSYM_PREFIX "dbd_get_seq_last")) == NULL) ||
@@ -925,7 +1083,7 @@ static dbi_driver_t *_get_driver(const char *filename) {
 		driver->functions->register_driver(&driver->info, &custom_functions_list, &driver->reserved_words);
 		driver->custom_functions = NULL; /* in case no custom functions are available */
 		while (custom_functions_list && custom_functions_list[idx] != NULL) {
-			custom = (dbi_custom_function_t *) malloc(sizeof(dbi_custom_function_t));
+			custom = malloc(sizeof(dbi_custom_function_t));
 			if (!custom) {
 				_free_custom_functions(driver);
 				free(driver->functions);
@@ -1036,7 +1194,7 @@ static dbi_option_t *_find_or_create_option_node(dbi_conn Conn, const char *key)
 
 	if (option == NULL) {
 		/* allocate a new option node */
-		option = (dbi_option_t *) malloc(sizeof(dbi_option_t));
+		option = malloc(sizeof(dbi_option_t));
 		if (!option) return NULL;
 		option->next = NULL;
 		option->key = strdup(key);
@@ -1052,8 +1210,10 @@ static dbi_option_t *_find_or_create_option_node(dbi_conn Conn, const char *key)
 	return option;
 }
 
+#define COUNTOF(array) (sizeof(array)/sizeof((array)[0]))
+
 void _error_handler(dbi_conn_t *conn, dbi_error_flag errflag) {
-	int errno = 0;
+	int my_errno = 0;
 	char *errmsg = NULL;
 	int errstatus;
 	static const char *errflag_messages[] = {
@@ -1066,7 +1226,8 @@ void _error_handler(dbi_conn_t *conn, dbi_error_flag errflag) {
 		/* DBI_ERROR_BADNAME */		"An invalid name was passed to libdbi",
 		/* DBI_ERROR_UNSUPPORTED */	"This particular libdbi driver or connection does not support this feature",
 		/* DBI_ERROR_NOCONN */		"libdbi could not establish a connection",
-		/* DBI_ERROR_NOMEM */		"libdbi ran out of memory" };
+		/* DBI_ERROR_NOMEM */		"libdbi ran out of memory",
+		/* DBI_ERROR_BADPTR */		"An invalid pointer was passed to libdbi"};
 	
 	if (conn == NULL) {
 		/* temp hack... if a result is disjoined and encounters an error, conn
@@ -1074,11 +1235,17 @@ void _error_handler(dbi_conn_t *conn, dbi_error_flag errflag) {
 		 * errors require a valid conn. this shouldn't even be a problem now,
 		 * since (currently) the only reason a result would be disjoint is if a
 		 * garbage collector was about to get rid of it. */
+		// conn will also be NULL if the connection itself could not be created,
+		// or if a NULL result handle was passed to a dbi routine, so we would
+		// always want to return if NULL. Probably should say something though:
+		fprintf(stderr, "libdbi: _error_handler: %s (NULL conn/result handle)\n",
+				errflag >= DBI_ERROR_BADOBJECT-1 && errflag < COUNTOF(errflag_messages)-1
+					? errflag_messages[errflag+1] : "");
 		return;
 	}
 
 	if (errflag == DBI_ERROR_DBD) {
-		errstatus = conn->driver->functions->geterror(conn, &errno, &errmsg);
+		errstatus = conn->driver->functions->geterror(conn, &my_errno, &errmsg);
 
 		if (errstatus == -1) {
 			/* not _really_ an error. XXX debug this, does it ever actually happen? */
@@ -1088,12 +1255,13 @@ void _error_handler(dbi_conn_t *conn, dbi_error_flag errflag) {
 
 	if (conn->error_message) free(conn->error_message);
 
-	if (errflag_messages[errflag+1] != NULL) {
+	if (errflag >= -1 && errflag < COUNTOF(errflag_messages)-1 
+			&& errflag_messages[errflag+1] != NULL) {
 		errmsg = strdup(errflag_messages[errflag+1]);
 	}
 	
 	conn->error_flag = errflag;
-	conn->error_number = errno;
+	conn->error_number = my_errno;
 	conn->error_message = errmsg;
 	
 	if (conn->error_handler != NULL) {
@@ -1102,17 +1270,87 @@ void _error_handler(dbi_conn_t *conn, dbi_error_flag errflag) {
 	}
 }
 
-unsigned long _isolate_attrib(unsigned long attribs, unsigned long rangemin, unsigned long rangemax) {
-	/* hahaha! who woulda ever thunk strawberry's code would come in handy? */
-	unsigned short startbit = log(rangemin)/log(2);
-	unsigned short endbit = log(rangemax)/log(2);
-	unsigned long attrib_mask = 0;
-	int x;
-	
-	for (x = startbit; x <= endbit; x++)
-		attrib_mask |= (1 << x);
+void _verbose_handler(dbi_conn_t *conn, const char* fmt, ...) {
+	va_list ap;
 
-	return (attribs & attrib_mask);
+	//if(conn && dbi_conn_get_option_numeric(conn, "Verbosity"))
+	{
+	  fputs("libdbi: ",stderr);
+	  va_start(ap, fmt);
+	  vfprintf(stderr, fmt, ap);
+	  va_end(ap);
+	}
+}
+
+void _logquery(dbi_conn_t *conn, const char* fmt, ...) {
+	va_list ap;
+
+	if(conn && dbi_conn_get_option_numeric(conn, "LogQueries")){
+	  fputs("libdbi: ", stderr);
+	  va_start(ap, fmt);
+	  vfprintf(stderr, fmt, ap);
+	  va_end(ap);
+	}
+}
+
+void _logquery_null(dbi_conn_t *conn, const char* statement, size_t st_length) {
+	if(conn && dbi_conn_get_option_numeric(conn, "LogQueries")){
+	  fputs("libdbi: [query_null] ", stderr);
+	  fwrite(statement, 1, st_length, stderr);
+	  fputc('\n', stderr);
+	}
+}
+
+unsigned int _isolate_attrib(unsigned int attribs, unsigned int rangemin, unsigned int rangemax) {
+	/* hahaha! who woulda ever thunk strawberry's code would come in handy? */
+	// find first (highest) bit set; methods not using FP can be found at 
+	// http://graphics.stanford.edu/~seander/bithacks.html#IntegerLogObvious
+	unsigned startbit = log(rangemin)/log(2);
+	unsigned endbit = log(rangemax)/log(2);
+	// construct mask from startbit to endbit inclusive
+	unsigned attrib_mask = ((1<<(endbit+1))-1) ^ ((1<<startbit)-1);
+	return attribs & attrib_mask;
+}
+
+/* computes an int from a [[[[A.]B.]C.]D.]E version number according to this
+  formula: version = E + D*100 + C*10000 + B*1000000 + A*100000000 */
+static unsigned int _parse_versioninfo(const char *version) {
+  char my_version[VERSIONSTRING_LENGTH];
+  char* start;
+  char* dot;
+  int i = 0;
+  unsigned int n_version = 0;
+  unsigned int n_multiplier = 1;
+
+  if (!version || !*version) {
+    return 0;
+  }
+
+  /* get a copy of version to mangle */
+  strncpy(my_version, version, VERSIONSTRING_LENGTH-1);
+  my_version[VERSIONSTRING_LENGTH-1] = '\0';
+
+  /* remove trailing periods */
+  if (my_version[strlen(my_version)-1] == '.') {
+    my_version[strlen(my_version)-1] = '\0';
+  }
+
+  start = my_version;
+  while ((dot = strrchr(start, (int)'.')) != NULL && i<5) {
+    n_version += atoi(dot+1) * n_multiplier;
+    *dot = '\0';
+    n_multiplier *= 100;
+    i++;
+  }
+  
+  /* take care of the remainder */
+  n_version += atoi(start) * n_multiplier;
+
+  if (i==5) {
+    /* version string can't be encoded in an unsigned int */
+    return 0;
+  }
+  return n_version;
 }
 
 #if HAVE_MACH_O_DYLD_H
@@ -1135,7 +1373,7 @@ static void * dyld_dlopen(const char * file)
 }
 static void * dyld_dlsym(void * hand, const char * name)
 {
-        NSSymbol * sym=NSLookupSymbolInModule((NSModule)hand, name);
+        NSSymbol sym=NSLookupSymbolInModule((NSModule)hand, name);
 	void * addr = NULL;
 	dyld_error_set=0;
 	if (sym) addr=(void*)NSAddressOfSymbol(sym);
@@ -1166,3 +1404,63 @@ static const char * dyld_dlerror()
 	return errstr;
 }
 #endif /* HAVE_MACH_O_DYLD_H */
+
+#ifdef __MINGW32__
+static char win_errstr[512];
+static int win_err_set = 0;
+void *win_dlopen(const char *filename, int mode)
+{
+  HINSTANCE handle;
+  handle = LoadLibrary(filename);
+#ifdef _WIN32
+  if (! handle) {
+    sprintf(win_errstr, "Error code 1(%d) while loading library %s", GetLastError(), filename);
+	win_err_set =1;
+    return NULL;
+  }
+#else
+  if ((UINT) handle < 32) {
+    sprintf(win_errstr, "error code 1(%d) while loading library %s", (UINT) handle, filename);
+	win_err_set =1;
+    return NULL;
+  }
+#endif
+  return (void *) handle;
+}
+
+void *win_dlsym(void *handle1, const char *symname)
+{
+  HMODULE handle = (HMODULE) handle1;
+  void *symaddr;
+  symaddr = (void *) GetProcAddress(handle, symname);
+  if (symaddr == NULL)
+  {
+    sprintf(win_errstr, "can not find symbol %s", symname);
+	win_err_set =2;
+  }
+  return symaddr;
+}
+
+int win_dlclose(void *handle1)
+{
+  HMODULE handle = (HMODULE) handle1;
+#ifdef _WIN32
+  if (FreeLibrary(handle))
+    return 0;
+  else {
+    sprintf(win_errstr, "error code 3(%d) while closing library", GetLastError());
+	win_err_set =3;
+    return -1;
+  }
+#else
+  FreeLibrary(handle);
+  return 0;
+#endif
+}
+
+char *win_dlerror()
+{
+  return win_errstr;
+}
+
+#endif
